----------------------------------------------------------------------------------------------------
--[[				NUMPADBAR
----------------------------------------------------------------------------------------------------

ABOUT

title:		NumPadBar
category:	Action Bars
notes:	Displays a numpad action bar.
author:	Shaktar of Stonemaul
date:		2007-05-26
version:	2.10

DESCRIPTION

Displays a numpad action bar. You can change its position, shape, or size.

FEATURES

* Offers a choice of five popular keyboards
* Offers a choice of three action bar sizes
* Switches bars on form/stance changes, if desired
* Hides the numpad or labels if desired
* Locks all action bars during combat
* Prevents shift-dragging actions out of numpad buttons during combat
* Prevents dragging numpad off screen
* Saves settings individually for each character
* Includes extensive help screens
* Compatible with OmniCC, a ui mod that displays cooldowns on buttons

INSTRUCTIONS

The background of the numpad is usually invisible but you can turn it gray by clicking on it. You can move the numpad or display the drop down menu even if the background is invisible. You can move the numpad by dragging the background. You can display the drop down menu by right-clicking the background. You can set various options with this menu such as changing the shape or size of the numpad, or hiding the numpad or labels.

You can drag actions and macros into and out of the numpad buttons. Action bars must be unlocked and you must be out of combat.

To display cooldowns on the buttons, you can use a cooldown mod such as OmniCC.

There are a limited number of slots to save actions in. If you choose not to use stance switching, warriors and druids will not have enough spare slots. Four extra slots are needed. The rightmost 4 slots in the bottom action bar #2 must be used to save numpad actions in. If you place actions in those slots, you will overwrite actions in the numpad.

Stance switching is supported for druids, rogues, and warriors, but only for stances/forms that have stance bars. Druids have stance bars for bear, cat, moonkin, and tree forms. Priests have a stance bar for shadowform. Rogues have a stance bar for stealth. Warriors have stance bars for battle, defensive, and berserker stances. The numeric keys on the numpad switch for these stances. Non-numeric keys remain the same. Stance switching can be turned off if desired.

NUMLOCK KEY

If your PC does not start up with your keyboard in NumLock mode, you can enable it in your BIOS. Restart your PC and (depending on your PC) either Del, F1, F2, F10, Esc, Ctl-S, or your PC's BIOS setup key before windows begins to load. Then set the NumLock option in BIOS to on and exit the BIOS.

Pressing the NumLock key while in NumLock mode changes the numpad to a cursor keypad. For example, shift-8 remaps to up-arrow and shift-1 remaps to End. If you accidentally press NumLock, then pressing numpad keys will no longer perform the actions you expect but will have completely different actions. You will have to press the NumLock key again to change it back to normal.

To avoid this problem permanently, you can use a keyboard remapping app (exe) such as KeyTweak to either disable or remap the NumLock button on your keyboard. At least disable this key so that this problem will never happen. Or better yet, remap it to the Escape key, so that you can use it as a Clear key when using calculator apps. You can use NumPadBar without remapping the NumLock key; just avoid pressing it.

Holding the shift key down while pressing numpad keys has the same effect as pressing the NumLock key and then pressing numpad keys. Unexpected actions will be performed. This problem cannot be avoided using a keyboard remapper. It can only be avoided by never holding the shift key when pressing numpad keys.

The NumLock key is used by WoW as the default key binding for Toggle Autorun. If you disable this key to avoid the problems mentioned above but still want a key for Toggle Autorun, you will have to set a new key binding for it.

KEY BINDINGS

All of the numeric/operator key bindings are used for the numpad. This overrides numpad-plus and numpad-minus for minimap zoom.

The Enter key can optionally be displayed in the numpad. If displayed, the Enter key binding is used for the numpad, overriding Open Chat. (The regular Enter key and the numpad Enter key are considered the same by WoW.) You will have to set a new key binding for the Open Chat function. The apostrophe key makes for a good key binding for Open Chat.

The Cursor keys can optionally be displayed. If displayed, the cursor and arrow key bindings are used for the numpad, overriding Previous Action Bar, Next Action Bar, movement (arrow) keys, chat scroll functions, and camera angle functions.

Both the numpad key bindings and the shifted numpad key bindings are used. So, for example, pressing numpad-plus results in the same action as pressing shift-numpad-plus. This helps prevent unexpected behavior especially during combat. 

If you hide the numpad or de-install NumPadBar, all of the normal WoW numpad key bindings will be restored to their original bindings.

You can still use the alt key for self-casting.

MOUSE BINDINGS

Previous Action Bar and Next Action Bar are bound by default to shift-mousewheelup and shift-mousewheeldown. When moving your hand from the numpad to the mouse, it is too easy to hit those bindings and accidentally scroll the bottom action bar.

To avoid this problem, you should unbind Previous Action Bar and Next Action Bar using the built-in WoW key bindings interface panel. You can rebind them to keys other than numpad keys.

SUMMARY

(1) Do not use the NumLock key (or remap it)
(2) Do not hold the shift key down when using the numpad
(3) If you choose to display the enter key: Rebind Open Chat to something other than Enter (apostrophe works)
(4) Unbind Previous Action Bar and Next Action Bar

CHANGE LOG

2.00	first version, numbered 2.0 to signify compatibility with WoW 2.0
2.01	a) code to check for dragging offscreen fixed, now can drag anwhere on screen at any resolution
	b) default keyboard now does not include enter & cursor keys, warnings about key bindings added
	c) warriors & druids no longer give up 2 action bars, just 4 rightmost slots in bottom bar #2
	d) more clear for first time users that the numpad can be left-clicked to drag and right-clicked for the menu
	e) menus simplified and clarified
	note) all users will have to rearrange their actions in the numpad once after upgrading
2.10	a) added stance switching for druids, priests, rogues, warriors (classes that have stance bars)
	a) 1) druid stances: bear, cat, moonkin, tree
	a) 2) priest stances: shadowform
	a) 3) rogue stances: stealth
	a) 4) warrior stances: battle, defensive, berserker
	b) added preference to turn off stance switching for those classes
	c) added enter key binding preference and cursor key binding preference
	d) no longer flashes the background for mouseover events
	e) dragging the numpad or using the menu cause the background to display momentarily
	f) clicking on the background toggles between displaying and hiding the background
	g) fixed bug where action bars would remain locked after combat
	note) all users will have to rearrange their actions in the numpad once after upgrading

]] -- edited with SciTE, best viewed using Verdana font and tabs and indents set to 8



----------------------------------------------------------------------------------------------------
-- DEFAULTS
----------------------------------------------------------------------------------------------------

-- the shape is determined by the brand of keyboard
NPB_DEFAULT_SHAPE = 'Windows'	-- should be the most common, defaults to no enter/cursor keys

-- the enter key is initially not used since it would unbind open chat
NPB_DEFAULT_ENTER = ''

-- the cursor keys are initially not used since they would unbind movement keys
NPB_DEFAULT_CURSOR = ''

-- size is 1 for normal action button size, or smaller for smaller buttons
-- for example, .8 would be 80% of normal size
NPB_DEFAULT_SIZE = 1			-- initially normal button size

-- save the position of the numpad frame
NPB_DEFAULT_LEFT = 900
NPB_DEFAULT_TOP = -500

-- labels are shortened versions of key bindings, example: NL for NumLock
NPB_DEFAULT_LABELS = true		-- display labels

-- some characters and alts may not want a numpad
NPB_DEFAULT_HIDDEN = false	-- initially not hidden

-- this is set to true after the mistakenly filled actions are cleared before creating numpads
NPB_DEFAULT_CLEARED = false

-- this is set to true after the user left-clicks or right-clicks on the numpad background
-- while this is false, messages tell the user that they can left-click or right-click on the background
NPB_DEFAULT_CLICKED = false

----------------------------------------------------------------------------------------------------
-- SAVED VARIABLES
----------------------------------------------------------------------------------------------------

NumPadBar = { }							-- defaults will be filled in by CheckSavedValues()
NumPadBar.cleared = NPB_DEFAULT_CLEARED	-- except for this default value
NumPadBar.leftclicked = NPB_DEFAULT_CLICKED	-- and this one
NumPadBar.rightclicked = NPB_DEFAULT_CLICKED	-- and this one

----------------------------------------------------------------------------------------------------
-- GLOBAL VARIABLES
----------------------------------------------------------------------------------------------------

NPB_numpads = { }			-- table of numpads (frames with buttons)
NPB_lockbars = nil				-- the original value for LOCK_ACTIONBAR
NPB_clickme = false				-- so that clickme message displayed only once per session

----------------------------------------------------------------------------------------------------
-- CONSTANTS
----------------------------------------------------------------------------------------------------

NPB_BACKDROP = {bgFile = 'Interface/Tooltips/UI-Tooltip-Background'}

NPB_BUTTON_SIZE = 42
NPB_DRAG_HANDLE = 16
NPB_DRAG_WIDTH = NPB_BUTTON_SIZE * 4 - 8
NPB_DRAG_HEIGHT = NPB_BUTTON_SIZE * 5 + 8

NPB_SIZES = {
	[1] = .8,
	[.8] = .67,
	[.67] = 1
	}

NPB_MENU = {
	windows = 'Keyboard is Windows',
	office = 'Keyboard is Microsoft Office',
	multimedia = 'Keyboard is Natural Multimedia',
	elite = 'Keyboard is Natural Elite',
	macintosh = 'Keyboard is Macintosh',
	medium = 'Size is Medium',
	small = 'Size is Small',
	tiny = 'Size is Tiny',
	enter = 'Toggle Enter Key',
	cursor = 'Toggle Cursor Keys',
	labels = 'Toggle Labels',
	stancebars = 'Toggle Stance Switching',
	hide = 'Hide Numpad',
	help = 'Help'
}

----------------------------------------------------------------------------------------------------
-- MESSAGES
----------------------------------------------------------------------------------------------------

NPB_PASTEL = '|cffffff80'		-- pastel yellow
NPB_BRIGHT = '|cffffff00'		-- bright yellow
NPB_PLAIN = '|r'			-- reset color

NPB_TITLE = NPB_BRIGHT .. 'NumPadBar 2.0 ' .. NPB_PASTEL .. 'by Shaktar of Stonemaul'

NPB_DESCRIPTION = NPB_BRIGHT .. 'NumPadBar 2.0: ' .. NPB_PASTEL .. 'Displays a numpad action bar. You can change its position, shape, or size. Move the numpad by dragging the background. Display the drop down menu by right-clicking the background. For instructions, select help from the menu or type /numpadbar help.'

NPB_LOADED = 'NumPadBar loaded. For info, type /numpadbar'
NPB_HIDDEN = 'NumPadBar is hidden. To show it, type /numpadbar'
NPB_RESET = 'All saved NumPadBar values are now reset to default.'

NPB_CLICK_ERROR = 'Left-Click to drag, Right-Click to display drop down menu'
NPB_ENTER_ERROR = 'Open Chat (enter key) now unbound!'
NPB_CURSOR_ERROR = 'Movement (arrow) keys now unbound!'

NPB_USING_STANCEBARS = 'The numeric keys and the decimal point key on the numpad will now switch to the appropriate stance bar on stance changes; keys 0-9 and [.] on the numpad will match keys 0-9 and [-] on the stance bar. Non-numeric keys on the numpad will always have the same action no matter what stance.'
NPB_NOT_USING_STANCEBARS = 'Keys on the numpad will no longer switch on stance changes.'
NPB_STANCELESS_CLASS = 'Only druids, rogues, and warriors have stance bars. Your class does not use stance bars.'

----------------------------------------------------------------------------------------------------
-- ACTION BAR SLOTS
----------------------------------------------------------------------------------------------------

-- 1 to 12		default action bar 1		bottom bar 1 (not used by warriors)
-- 13 to 24		default action bar 2		bottom bar 2
-- 25 to 36		default action bar 3		right action bar (rightmost of the 2 right bars)
-- 37 to 48		default action bar 4		right action bar 2 (leftmost of the 2 right bars)
-- 49 to 60		default action bar 5		lower right bar
-- 61 to 72		default action bar 6		lower left bar
-- 73 to 84		warrior battle stance, druid cat form, rogue stealth, priest shadowform
-- 85 to 96		warrior defensive stance, druid tree form
-- 97 to 108	warrior berserker stance, druid bear form
-- 109 to 120	unused action slots
-- unknown: druid moonkin form

NPB_NORMAL_SLOTS = {
	85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96,
	97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108,
	109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120
}

NPB_DRUID_SLOTS = {
	1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
	109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
	24, 23, 22, 21
}

NPB_WARRIOR_SLOTS = {
	1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
	109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
	24, 23, 22, 21
}

NPB_CLASS_SLOTS = {
	druid = NPB_DRUID_SLOTS,
	hunter = NPB_NORMAL_SLOTS,
	mage = NPB_NORMAL_SLOTS,
	paladin = NPB_NORMAL_SLOTS,
	priest = NPB_NORMAL_SLOTS,
	rogue = NPB_NORMAL_SLOTS,
	shaman = NPB_NORMAL_SLOTS,
	warlock = NPB_NORMAL_SLOTS,
	warrior = NPB_WARRIOR_SLOTS
}

----------------------------------------------------------------------------------------------------
-- STANCE NUMBERS
----------------------------------------------------------------------------------------------------

--[[
Class:		Warrior	Priest	Rogue	Druid
-------------------------------------------------------------------------------------
Stance:1	Battle	Shadow	Stealth	Bear
Stance:2	Defensive					Aquatic
Stance:3	Berserker					Cat
Stance:4							Travel
Stance:5							Moonkin/Tree/Flight*
Stance:6							Flight
*If you have both e.g. Moonkin form and Flying form then Moonkin is 5 and Flight is 6.
]]

----------------------------------------------------------------------------------------------------
-- STANCE BARS
-- see http://www.wowwiki.com/API_GetShapeshiftForm
----------------------------------------------------------------------------------------------------

NPB_DRUID_STANCEBARS = {
	[0] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 },						-- humanoid
	[1] = { 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108 },	-- bear
	[3] = { 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84 },				-- cat
	[5] = { 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96 },				-- moonkin?/tree
	[9] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 }						-- non-stanced
}

NPB_PRIEST_STANCEBARS = {
	[0] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 },						-- humanoid
	[1] = { 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84 },				-- shadowform
	[9] = { 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96 }				-- non-stanced
}

NPB_ROGUE_STANCEBARS = {
	[0] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 },						-- humanoid
	[1] = { 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108 },	-- stealth
	[9] = { 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96 }				-- non-stanced
}

NPB_WARRIOR_STANCEBARS = {
	[1] = { 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84 },				-- battle
	[2] = { 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96 },				-- defensive
	[3] = { 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108 },	-- berserker
	[9] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 }						-- non-stanced
}

NPB_CLASS_STANCEBARS = {
	druid = NPB_DRUID_STANCEBARS,
	priest = NPB_PRIEST_STANCEBARS,
	rogue = NPB_ROGUE_STANCEBARS,
	warrior = NPB_WARRIOR_STANCEBARS
}

NPB_CLASS_STATES = {
	druid = '0:S0;1:S1;2:S0;3:S3;4:S0;5:S5;6:S0;',
	priest = '0:S0;1:S1;',
	rogue = '0:S0;1:S1;',
	warrior = '1:S1;2:S2;3:S3;'
}

----------------------------------------------------------------------------------------------------
-- KEYBOARD LAYOUTS
----------------------------------------------------------------------------------------------------

NPB_KEYBOARDS = {

Windows = {

	{ row = 4, col = 1, type='Numeric', label = '1', binding = 'NumPad1' },
	{ row = 4, col = 2, type='Numeric', label = '2', binding = 'NumPad2' },
	{ row = 4, col = 3, type='Numeric', label = '3', binding = 'NumPad3' },
	{ row = 3, col = 1, type='Numeric', label = '4', binding = 'NumPad4' },
	{ row = 3, col = 2, type='Numeric', label = '5', binding = 'NumPad5' },
	{ row = 3, col = 3, type='Numeric', label = '6', binding = 'NumPad6' },
	{ row = 2, col = 1, type='Numeric', label = '7', binding = 'NumPad7' },
	{ row = 2, col = 2, type='Numeric', label = '8', binding = 'NumPad8' },
	{ row = 2, col = 3, type='Numeric', label = '9', binding = 'NumPad9' },
	{ row = 5, col = 1.5, type='Numeric', label = '0', binding = 'NumPad0' },

	{ row = 5, col = 3, type='Numeric', label = '.', binding = 'NumPadDecimal' },
	{ row = 2.5, col = 4, type='Numeric', label = '+', binding = 'NumPadPlus' },
	{ row = 1, col = 4, type='Numeric', label = '-', binding = 'NumPadMinus' },
	{ row = 1, col = 3, type='Numeric', label = '*', binding = 'NumPadMultiply' },
	{ row = 1, col = 2, type='Numeric', label = '/', binding = 'NumPadDivide' },

	{ row = 4.5, col = 4, type='Enter', label = 'E', binding = 'Enter' },

	{ row = 4, col = 2 - 3.3, type='Cursor', label = '^', binding = 'Up' },
	{ row = 5, col = 1 - 3.3, type='Cursor', label = '<', binding = 'Left' },
	{ row = 5, col = 2 - 3.3, type='Cursor', label = 'v', binding = 'Down' },
	{ row = 5, col = 3 - 3.3, type='Cursor', label = '>', binding = 'Right' },
	
	{ row = 1, col = 1 - 3.3, type='Cursor', label = 'I', binding = 'Insert' },
	{ row = 1, col = 2 - 3.3, type='Cursor', label = 'H', binding = 'Home' },
	{ row = 1, col = 3 - 3.3, type='Cursor', label = 'P^', binding = 'PageUp' },
	{ row = 2, col = 1 - 3.3, type='Cursor', label = 'D', binding = 'Delete' },
	{ row = 2, col = 2 - 3.3, type='Cursor', label = 'E', binding = 'End' },
	{ row = 2, col = 3 - 3.3, type='Cursor', label = 'Pv', binding = 'PageDown' }

},

MicrosoftOffice = {

	{ row = 4, col = 1, type='Numeric', label = '1', binding = 'NumPad1' },
	{ row = 4, col = 2, type='Numeric', label = '2', binding = 'NumPad2' },
	{ row = 4, col = 3, type='Numeric', label = '3', binding = 'NumPad3' },
	{ row = 3, col = 1, type='Numeric', label = '4', binding = 'NumPad4' },
	{ row = 3, col = 2, type='Numeric', label = '5', binding = 'NumPad5' },
	{ row = 3, col = 3, type='Numeric', label = '6', binding = 'NumPad6' },
	{ row = 2, col = 1, type='Numeric', label = '7', binding = 'NumPad7' },
	{ row = 2, col = 2, type='Numeric', label = '8', binding = 'NumPad8' },
	{ row = 2, col = 3, type='Numeric', label = '9', binding = 'NumPad9' },
	{ row = 5, col = 1.5, type='Numeric', label = '0', binding = 'NumPad0' },

	{ row = 5, col = 3, type='Numeric', label = '.', binding = 'NumPadDecimal' },
	{ row = 2.5, col = 4, type='Numeric', label = '+', binding = 'NumPadPlus' },
	{ row = 1, col = 4, type='Numeric', label = '-', binding = 'NumPadMinus' },
	{ row = 1, col = 3, type='Numeric', label = '*', binding = 'NumPadMultiply' },
	{ row = 1, col = 2, type='Numeric', label = '/', binding = 'NumPadDivide' },

	{ row = 4.5, col = 4, type='Enter', label = 'E', binding = 'Enter' },

	{ row = 4.3, col = 2 - 3.3, type='Cursor', label = '^', binding = 'Up' },
	{ row = 5.3, col = 1 - 3.3, type='Cursor', label = '<', binding = 'Left' },
	{ row = 5.3, col = 2 - 3.3, type='Cursor', label = 'v', binding = 'Down' },
	{ row = 5.3, col = 3 - 3.3, type='Cursor', label = '>', binding = 'Right' },
	
	{ row = 1, col = 2 - 3.3, type='Cursor', label = 'H', binding = 'Home' },
	{ row = 1, col = 3 - 3.3, type='Cursor', label = 'E', binding = 'End' },
	{ row = 2, col = 3 - 3.3, type='Cursor', label = 'P^', binding = 'PageUp' },
	{ row = 2.5, col = 2 - 3.3, type='Cursor', label = 'D', binding = 'Delete' },
	{ row = 3, col = 3 - 3.3, type='Cursor', label = 'Pv', binding = 'PageDown' },

	-- the following key is extra on a windows office keyboard
	{ row = 1, col = 1, type='Cursor', label = 'T', binding = 'Tab' }

},

NaturalMultimedia = {

	{ row = 4, col = 1, type='Numeric', label = '1', binding = 'NumPad1' },
	{ row = 4, col = 2, type='Numeric', label = '2', binding = 'NumPad2' },
	{ row = 4, col = 3, type='Numeric', label = '3', binding = 'NumPad3' },
	{ row = 3, col = 1, type='Numeric', label = '4', binding = 'NumPad4' },
	{ row = 3, col = 2, type='Numeric', label = '5', binding = 'NumPad5' },
	{ row = 3, col = 3, type='Numeric', label = '6', binding = 'NumPad6' },
	{ row = 2, col = 1, type='Numeric', label = '7', binding = 'NumPad7' },
	{ row = 2, col = 2, type='Numeric', label = '8', binding = 'NumPad8' },
	{ row = 2, col = 3, type='Numeric', label = '9', binding = 'NumPad9' },
	{ row = 5, col = 1.5, type='Numeric', label = '0', binding = 'NumPad0' },

	{ row = 5, col = 3, type='Numeric', label = '.', binding = 'NumPadDecimal' },
	{ row = 2.5, col = 4, type='Numeric', label = '+', binding = 'NumPadPlus' },
	{ row = 1, col = 4, type='Numeric', label = '-', binding = 'NumPadMinus' },
	{ row = 1, col = 3, type='Numeric', label = '*', binding = 'NumPadMultiply' },
	{ row = 1, col = 2, type='Numeric', label = '/', binding = 'NumPadDivide' },

	{ row = 4.5, col = 4, type='Enter', label = 'E', binding = 'Enter' },

	{ row = 4.3, col = 2 - 3.3, type='Cursor', label = '^', binding = 'Up' },
	{ row = 5.3, col = 1 - 3.3, type='Cursor', label = '<', binding = 'Left' },
	{ row = 5.3, col = 2 - 3.3, type='Cursor', label = 'v', binding = 'Down' },
	{ row = 5.3, col = 3 - 3.3, type='Cursor', label = '>', binding = 'Right' },
	
	{ row = 1, col = 2 - 3.3, type='Cursor', label = 'H', binding = 'Home' },
	{ row = 1, col = 3 - 3.3, type='Cursor', label = 'E', binding = 'End' },
	{ row = 2, col = 3 - 3.3, type='Cursor', label = 'P^', binding = 'PageUp' },
	{ row = 2.5, col = 2 - 3.3, type='Cursor', label = 'D', binding = 'Delete' },
	{ row = 3, col = 3 - 3.3, type='Cursor', label = 'Pv', binding = 'PageDown' }

},

NaturalElite = {

	{ row = 4, col = 1, type='Numeric', label = '1', binding = 'NumPad1' },
	{ row = 4, col = 2, type='Numeric', label = '2', binding = 'NumPad2' },
	{ row = 4, col = 3, type='Numeric', label = '3', binding = 'NumPad3' },
	{ row = 3, col = 1, type='Numeric', label = '4', binding = 'NumPad4' },
	{ row = 3, col = 2, type='Numeric', label = '5', binding = 'NumPad5' },
	{ row = 3, col = 3, type='Numeric', label = '6', binding = 'NumPad6' },
	{ row = 2, col = 1, type='Numeric', label = '7', binding = 'NumPad7' },
	{ row = 2, col = 2, type='Numeric', label = '8', binding = 'NumPad8' },
	{ row = 2, col = 3, type='Numeric', label = '9', binding = 'NumPad9' },
	{ row = 5, col = 1.5, type='Numeric', label = '0', binding = 'NumPad0' },

	{ row = 5, col = 3, type='Numeric', label = '.', binding = 'NumPadDecimal' },
	{ row = 2.5, col = 4, type='Numeric', label = '+', binding = 'NumPadPlus' },
	{ row = 1, col = 4, type='Numeric', label = '-', binding = 'NumPadMinus' },
	{ row = 1, col = 3, type='Numeric', label = '*', binding = 'NumPadMultiply' },
	{ row = 1, col = 2, type='Numeric', label = '/', binding = 'NumPadDivide' },

	{ row = 4.5, col = 4, type='Enter', label = 'E', binding = 'Enter' },

	{ row = 4.3, col = 2 - 3.1, type='Cursor', label = '^', binding = 'Up' },
	{ row = 4.8, col = 1 - 3.1, type='Cursor', label = '<', binding = 'Left' },
	{ row = 5.3, col = 2 - 3.1, type='Cursor', label = 'v', binding = 'Down' },
	{ row = 4.8, col = 3 - 3.1, type='Cursor', label = '>', binding = 'Right' },

	{ row = 1, col = 1 - 2.6, type='Cursor', label = 'H', binding = 'Home' },
	{ row = 1, col = 2 - 2.6, type='Cursor', label = 'P^', binding = 'PageUp' },
	{ row = 2, col = 1 - 2.6, type='Cursor', label = 'E', binding = 'End' },
	{ row = 2, col = 2 - 2.6, type='Cursor', label = 'Pv', binding = 'PageDown' },
	{ row = 3, col = 1 - 2.6, type='Cursor', label = 'D', binding = 'Delete' },
	{ row = 3, col = 2 - 2.6, type='Cursor', label = 'I', binding = 'Insert' }

},
	
Macintosh = {

	{ row = 4, col = 1, type='Numeric', label = '1', binding = 'NumPad1' },
	{ row = 4, col = 2, type='Numeric', label = '2', binding = 'NumPad2' },
	{ row = 4, col = 3, type='Numeric', label = '3', binding = 'NumPad3' },
	{ row = 3, col = 1, type='Numeric', label = '4', binding = 'NumPad4' },
	{ row = 3, col = 2, type='Numeric', label = '5', binding = 'NumPad5' },
	{ row = 3, col = 3, type='Numeric', label = '6', binding = 'NumPad6' },
	{ row = 2, col = 1, type='Numeric', label = '7', binding = 'NumPad7' },
	{ row = 2, col = 2, type='Numeric', label = '8', binding = 'NumPad8' },
	{ row = 2, col = 3, type='Numeric', label = '9', binding = 'NumPad9' },
	{ row = 5, col = 1.5, type='Numeric', label = '0', binding = 'NumPad0' },

	{ row = 5, col = 3, type='Numeric', label = '.', binding = 'NumPadDecimal' },
	{ row = 3, col = 4, type='Numeric', label = '+', binding = 'NumPadPlus' },
	{ row = 2, col = 4, type='Numeric', label = '-', binding = 'NumPadMinus' },
	{ row = 1, col = 4, type='Numeric', label = '*', binding = 'NumPadMultiply' },
	{ row = 1, col = 3, type='Numeric', label = '/', binding = 'NumPadDivide' },
	
	{ row = 4.5, col = 4, type='Enter', label = 'E', binding = 'Enter' },
	
	{ row = 4, col = 2 - 3.3, type='Cursor', label = '^', binding = 'Up' },
	{ row = 5, col = 1 - 3.3, type='Cursor', label = '<', binding = 'Left' },
	{ row = 5, col = 2 - 3.3, type='Cursor', label = 'v', binding = 'Down' },
	{ row = 5, col = 3 - 3.3, type='Cursor', label = '>', binding = 'Right' },

	{ row = 1, col = 1 - 3.3, type='Cursor', label = 'H?', binding = 'Insert' },   -- Help
	{ row = 1, col = 2 - 3.3, type='Cursor', label = 'H', binding = 'Home' },
	{ row = 1, col = 3 - 3.3, type='Cursor', label = 'P^', binding = 'PageUp' },
	{ row = 2, col = 1 - 3.3, type='Cursor', label = 'D', binding = 'Delete' },
	{ row = 2, col = 2 - 3.3, type='Cursor', label = 'E', binding = 'End' },
	{ row = 2, col = 3 - 3.3, type='Cursor', label = 'Pv', binding = 'PageDown' },

	-- the following 2 keys are extra on a macintosh
	-- + and - key on mac take up same room as + key on pc
	{ row = 1, col = 2, type='Numeric', label = '=', binding = 'NumPadEquals' },
	{ row = 1, col = 1, type='Cursor', label = 'C', binding = 'NumLock' }   -- Clear

}

}

----------------------------------------------------------------------------------------------------
-- INSTRUCTIONS
----------------------------------------------------------------------------------------------------

NPB_INSTRUCTIONS =

NPB_TITLE .. '\r\r' .. NPB_PLAIN ..

[[
The background of the numpad is usually invisible but you can turn it gray by clicking on it. You can move the numpad or display the drop down menu even if the background is invisible. You can move the numpad by dragging the background. You can display the drop down menu by right-clicking the background. You can set various options with this menu such as changing the shape or size of the numpad, or hiding the numpad or labels.

You can drag actions and macros into and out of the numpad buttons. Action bars must be unlocked and you must be out of combat.

To display cooldowns on the buttons, you can use a cooldown mod such as OmniCC.

There are a limited number of slots to save actions in. If you choose not to use stance switching, warriors and druids will not have enough spare slots. Four extra slots are needed. The rightmost 4 slots in the bottom action bar #2 must be used to save numpad actions in. If you place actions in those slots, you will overwrite actions in the numpad.

Stance switching is supported for druids, rogues, and warriors, but only for stances/forms that have stance bars. Druids have stance bars for bear, cat, moonkin, and tree forms. Priests have a stance bar for shadowform. Rogues have a stance bar for stealth. Warriors have stance bars for battle, defensive, and berserker stances. The numeric keys on the numpad switch for these stances. Non-numeric keys remain the same. Stance switching can be turned off if desired.
]]

if (not StaticPopupDialogs) then StaticPopupDialogs = { } end   -- to make SciTE happy
StaticPopupDialogs['NPB_INSTRUCTIONS'] = {
	text = NPB_INSTRUCTIONS,
	button1 = 'OK',
	timeout = 0,
	whileDead = 1,
	hideOnEscape = 1,
	OnHide = function() StaticPopup_Show('NPB_NUM_LOCK') end
}

----------------------------------------------------------------------------------------------------

NPB_NUM_LOCK = NPB_BRIGHT .. 'NumLock Key\r\r' .. NPB_PLAIN ..

[[
If your PC does not start up with your keyboard in NumLock mode, you can enable it in your BIOS. Restart your PC and (depending on your PC) either Del, F1, F2, F10, Esc, Ctl-S, or your PC's BIOS setup key before windows begins to load. Then set the NumLock option in BIOS to on and exit the BIOS.

Pressing the NumLock key while in NumLock mode changes the numpad to a cursor keypad. For example, shift-8 remaps to up-arrow and shift-1 remaps to End. If you accidentally press NumLock, then pressing numpad keys will no longer perform the actions you expect but will have completely different actions. You will have to press the NumLock key again to change it back to normal.

To avoid this problem permanently, you can use a keyboard remapping app (exe) such as KeyTweak to either disable or remap the NumLock button on your keyboard. At least disable this key so that this problem will never happen. Or better yet, remap it to the Escape key, so that you can use it as a Clear key when using calculator apps. You can use NumPadBar without remapping the NumLock key; just avoid pressing it.

Holding the shift key down while pressing numpad keys has the same effect as pressing the NumLock key and then pressing numpad keys. Unexpected actions will be performed. This problem cannot be avoided using a keyboard remapper. It can only be avoided by never holding the shift key when pressing numpad keys.

The NumLock key is used by WoW as the default key binding for Toggle Autorun. If you disable this key to avoid the problems mentioned above but still want a key for Toggle Autorun, you will have to set a new key binding for it.
]]

StaticPopupDialogs['NPB_NUM_LOCK'] = {
	text = NPB_NUM_LOCK,
	button1 = 'OK',
	timeout = 0,
	whileDead = 1,
	hideOnEscape = 1,
	OnHide = function() StaticPopup_Show('NPB_KEY_BINDINGS') end
}

----------------------------------------------------------------------------------------------------

NPB_KEY_BINDINGS = NPB_BRIGHT .. 'Key Bindings\r\r' .. NPB_PLAIN ..

[[
All of the numeric/operator key bindings are used for the numpad. This overrides numpad-plus and numpad-minus for minimap zoom.

The Enter key can optionally be displayed in the numpad. If displayed, the Enter key binding is used for the numpad, overriding Open Chat. (The regular Enter key and the numpad Enter key are considered the same by WoW.) You will have to set a new key binding for the Open Chat function. The apostrophe key makes for a good key binding for Open Chat.

The Cursor keys can optionally be displayed. If displayed, the cursor and arrow key bindings are used for the numpad, overriding Previous Action Bar, Next Action Bar, movement (arrow) keys, chat scroll functions, and camera angle functions.

Both the numpad key bindings and the shifted numpad key bindings are used. So, for example, pressing numpad-plus results in the same action as pressing shift-numpad-plus. This helps prevent unexpected behavior especially during combat. 

If you hide the numpad or de-install NumPadBar, all of the normal WoW numpad key bindings will be restored to their original bindings.

You can still use the alt key for self-casting.
]]

StaticPopupDialogs['NPB_KEY_BINDINGS'] = {
	text = NPB_KEY_BINDINGS,
	button1 = 'OK',
	timeout = 0,
	whileDead = 1,
	hideOnEscape = 1,
	OnHide = function() StaticPopup_Show('NPB_MOUSE_BINDINGS') end
}

----------------------------------------------------------------------------------------------------

NPB_MOUSE_BINDINGS = NPB_BRIGHT .. 'Mouse Bindings\r\r' .. NPB_PLAIN ..

[[
Previous Action Bar and Next Action Bar are bound by default to shift-mousewheelup and shift-mousewheeldown. When moving your hand from the numpad to the mouse, it is too easy to hit those bindings and accidentally scroll the bottom action bar.

To avoid this problem, you should unbind Previous Action Bar and Next Action Bar using the built-in WoW key bindings interface panel. You can rebind them to keys other than numpad keys.
]]

.. NPB_BRIGHT .. '\rSummary\r\r' .. NPB_PLAIN ..

[[
(1) Do not use the NumLock key (or remap it)
(2) Do not hold the shift key down when using the numpad
(3) If you choose to display the enter key: Rebind Open Chat to something other than Enter (apostrophe works)
(4) Unbind Previous Action Bar and Next Action Bar
]]

StaticPopupDialogs['NPB_MOUSE_BINDINGS'] = {
	text = NPB_MOUSE_BINDINGS,
	button1 = 'OK',
	timeout = 0,
	whileDead = 1,
	hideOnEscape = 1
}

----------------------------------------------------------------------------------------------------
-- DEBUGGING FUNCTIONS
-- two ways to use bug() function
-- first way is to pass it a list of arguments, it displays the values of all arguments
-- second way is to pass it a single table, it displays all the keys and values in the table
----------------------------------------------------------------------------------------------------

function bug(...)
if (false) then return end
local msg = ''
local isTable = false
if (select('#', ...) == 1 and type(select(1, ...)) == 'table') then
	isTable = true end
if (isTable) then
	for key, value in pairs(select(1, ...)) do
		msg = msg .. tostring(key) .. '=' .. tostring(value)
		if (next(select(1, ...), key)) then msg = msg .. ', ' end
		end
	end
if (not isTable) then
	for ix = 1, select('#', ...) do
		if (ix > 1) then msg = msg .. ', ' end
		msg = msg .. tostring(select(ix, ...))
		end
	end
msg = msg .. '::'
DEFAULT_CHAT_FRAME:AddMessage(msg)
end

----------------------------------------------------------------------------------------------------
-- WoW FUNCTIONS
----------------------------------------------------------------------------------------------------

function ActionButton_UpdateHotkeys(actionButtonType)
    if (getglobal(this:GetName()..'HotKey'):GetText()) then return end   -- /shaktar
	-- if already has a label, do nothing   /shaktar
    if ( not actionButtonType ) then
        actionButtonType = "ACTIONBUTTON";
    end
    local hotkey = getglobal(this:GetName().."HotKey");
    local button = ActionButton_GetPagedID(this);
    local key = GetBindingKey(actionButtonType..this:GetID()) or
                GetBindingKey("CLICK "..this:GetName()..":LeftButton");

    if ( GetBindingText(key, "KEY_", 1) == "" ) then
        if ( not HasAction(button) ) then
            hotkey:SetText("");
        elseif ( ActionHasRange(button) ) then
            if ( IsActionInRange(button) ) then
                hotkey:SetText(RANGE_INDICATOR);
                hotkey:SetTextHeight(8);
                hotkey:SetPoint("TOPRIGHT", this:GetName(), "TOPRIGHT", -3, 5);
            else
                hotkey:SetText("");
            end
        end
    else
        hotkey:SetText(GetBindingText(key, "KEY_", 1));
    end
end

----------------------------------------------------------------------------------------------------
-- NUMPADBAR FUNCTIONS
----------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------
function NPB_time()
-- returns system time in seconds
----------------------------------------------------------------------------------------------------
if (GetTime) then return GetTime() end
if (os and os.time) then return os.time() end
return 0
end

----------------------------------------------------------------------------------------------------
function NPB_Chat(message)
----------------------------------------------------------------------------------------------------

-- displays message in the chat frame
if (not message or not DEFAULT_CHAT_FRAME) then return end
DEFAULT_CHAT_FRAME:AddMessage(message)
end

----------------------------------------------------------------------------------------------------
function NPB_DoSlash(cmd)
----------------------------------------------------------------------------------------------------

cmd = string.lower(cmd)

if (cmd == 'help') then
	StaticPopup_Show('NPB_INSTRUCTIONS')
	return end

-- displays the key bound to the Jump action in a string that can be used in code
-- this may not be the same as the string displayed to the user in the keybindings dialog
-- this slash command is only useful for development, to get keybinding strings for various keys
if (cmd == 'jump') then
	NPB_Chat(GetBindingKey('Jump'))
	return end

-- for development purposes, lists all of the used slots
if (cmd == 'slots') then
	NPB_ListSlots()
	return end

-- for development purposes, resets all values to default
if (cmd == 'reset') then
	NPB_HideNumpad()
	NumPadBar = { }
	NPB_clickme = false
	NPB_CheckValues()
	NPB_ShowNumpad()
	NPB_Chat(NPB_RESET)
	return end

local numpad = NPB_GetNumpad()
SetBackdrop(numpad)
NPB_Chat(NPB_DESCRIPTION)
NPB_ShowNumpad()
end

----------------------------------------------------------------------------------------------------
function NPB_ListSlots()
----------------------------------------------------------------------------------------------------

for slot = 1, 120 do
	local texture = GetActionTexture(slot)
	if texture then
		NPB_Chat(slot .. ': ' .. texture) end
	end
end

----------------------------------------------------------------------------------------------------
function NPB_DoLoad()
----------------------------------------------------------------------------------------------------

this:RegisterEvent('PLAYER_LOGIN')
this:RegisterEvent('PLAYER_REGEN_DISABLED')
this:RegisterEvent('PLAYER_REGEN_ENABLED')
SlashCmdList['NPB'] = NPB_DoSlash
SLASH_NPB1 = '/numpadbar'
end

----------------------------------------------------------------------------------------------------
function NPB_DoEvent(event)
----------------------------------------------------------------------------------------------------

if (event == 'PLAYER_LOGIN') then
	NPB_Initialize() end

-- stop event handling if numpad is hidden
if (NumPadBar.hidden) then return end

-- in combat, so lock action bars
if (event == 'PLAYER_REGEN_DISABLED') then
	NPB_LockBars() end

-- out of combat, so unlock bars
if (event == 'PLAYER_REGEN_ENABLED') then
	NPB_UnlockBars() end

end

----------------------------------------------------------------------------------------------------
function NPB_Initialize()
----------------------------------------------------------------------------------------------------

local localizedClass, englishClass = UnitClass('player')
local class = string.lower(englishClass)
NPB_CheckValues()
NPB_ClearSlots(class)
if (NumPadBar.hidden) then NPB_Chat(NPB_HIDDEN) end
if (not NumPadBar.hidden) then
	NPB_ShowNumpad() end
end

----------------------------------------------------------------------------------------------------
function NPB_LockBars()
----------------------------------------------------------------------------------------------------

if (LOCK_ACTIONBAR == '1') then return end
local numpad = NPB_GetNumpad()
ClearBackdrop(numpad)		-- remove gray bg, don't restore it when out of combat
LOCK_ACTIONBAR = '1'
NumPadBar.unlock = true
end

----------------------------------------------------------------------------------------------------
function NPB_UnlockBars()
----------------------------------------------------------------------------------------------------

if (not NumPadBar.unlock) then return end
LOCK_ACTIONBAR = '0'
NumPadBar.unlock = false
end

----------------------------------------------------------------------------------------------------
function NPB_ClearSlots(class)
-- a couple of slots are filled with actions for certain classes, a blizzard bug
-- for warriors, slots 1 and 109 are mistakenly prefilled
-- for druids, slots 85 and 109 are mistakenly prefilled
-- those actions are cleared since the slots are used by numpadbar
-- this only happens once per character, before the numpads are created
----------------------------------------------------------------------------------------------------

if (NumPadBar.cleared) then return end

if (class == 'warrior' and (GetActionTexture(1) or GetActionTexture(109))) then
	PickupAction(1)
	PutItemInBackpack()
	PickupAction(109)
	PutItemInBackpack()
	end

if (class == 'druid' and (GetActionTexture(85) or GetActionTexture(109))) then
	PickupAction(85)
	PutItemInBackpack()
	PickupAction(109)
	PutItemInBackpack()
	end

NumPadBar.cleared = true
end

----------------------------------------------------------------------------------------------------
function NPB_CheckValues()
-- checks for valid values before creating numpad frames
----------------------------------------------------------------------------------------------------

local localizedClass, englishClass = UnitClass('player')
local class = string.lower(englishClass)
local stances = NPB_CLASS_STANCEBARS[class]
local n = NumPadBar
if (not n) then
	n = { } end
if (not n.shape or not NPB_KEYBOARDS[n.shape]) then
	n.shape = NPB_DEFAULT_SHAPE end
if (not n.enter) then
	n.enter = '' end
if (not n.cursor) then
	n.cursor = '' end
if (not n.size or not NPB_SIZES[n.size]) then
	n.size = NPB_DEFAULT_SIZE end
if (not n.left or n.left < -500 or n.left > 2000) then
	n.left = NPB_DEFAULT_LEFT end
if (not n.top or n.top < -2000 or n.top > 500) then
	n.top = NPB_DEFAULT_TOP end
if (n.stancebars ~= false) then
	n.stancebars = true end
if (not stances) then
	n.stancebars = false end
if (n.labels ~= false) then
	n.labels = true end
end

----------------------------------------------------------------------------------------------------
function NPB_ClickMe()
-- displays a message in the error frame (upper screen) for new users/characters
-- the message talks about left-clicking and right-clicking the numpad
----------------------------------------------------------------------------------------------------

if (NumPadBar.leftclicked and NumPadBar.rightclicked) then return end
if (UnitAffectingCombat('player') == 1) then return end
if (NPB_clickme) then return end
NPB_clickme = true
NPB_Chat(NPB_DESCRIPTION)
UIErrorsFrame:AddMessage(NPB_CLICK_ERROR, 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME)
end

----------------------------------------------------------------------------------------------------
function NPB_SetState(numpad)
----------------------------------------------------------------------------------------------------

if (NumPadBar.stancebars) then
	local stance = GetShapeshiftForm(true)
	numpad:SetAttribute('statemap-stance', '$input')		-- map: stance ==> state
	numpad:SetAttribute('state', stance)				-- state = stance
else
	numpad:SetAttribute('statemap-stance', '')
	numpad:SetAttribute('state', 9)		-- so-called normal state, non-stance slots
	end
end

----------------------------------------------------------------------------------------------------
function NPB_CreateNumpad(class, shape, enter, cursor)
-- shape of numpad dictates the key layout
-- makes draggable frame containing buttons, adds it to a global table of numpads
-- does not drag numpad or display drop down menu or change numpad if in combat
----------------------------------------------------------------------------------------------------
-- local functions
----------------------------------------------------------------------------------------------------

	function NPB_DoDragStart()
	if (UnitAffectingCombat('player') == 1) then return end
	NumPadBar.leftclicked = true
	SetBackdrop(this)
	this:StartMoving()
	end

	function NPB_DoDragStop()
	this:StopMovingOrSizing()
	ClearBackdrop(this)
	local point, relativeTo, relativePoint, left, top = this:GetPoint()
	NumPadBar.left = left   -- save these values so that numpad is placed in same position next login
	NumPadBar.top = top
	end

	function NPB_DoMouseDown()
	if (UnitAffectingCombat('player') == 1) then return end
	if (IsMouseButtonDown('leftbutton')) then
		ToggleBackdrop(this)
	else
		SetBackdrop(this)
		NumPadBar.rightclicked = true
		NPB_ToggleMenu()
		end
	end

	function NPB_DoEnter()
	NPB_ClickMe()
	end

----------------------------------------------------------------------------------------------------
-- main function
----------------------------------------------------------------------------------------------------

local numpad = CreateFrame('Frame', 'NumpadFrame', UIParent, 'SecureStateDriverTemplate')
numpad:SetMovable(1)
numpad:EnableMouse(1)
numpad:RegisterForDrag('LeftButton')
numpad:SetScript('OnDragStart', NPB_DoDragStart)
numpad:SetScript('OnDragStop', NPB_DoDragStop)
numpad:SetScript('OnMouseDown', NPB_DoMouseDown)
numpad:SetScript('OnEnter', NPB_DoEnter)
numpad:SetWidth(NPB_DRAG_WIDTH)
numpad:SetHeight(NPB_DRAG_HEIGHT)
numpad:SetScale(NumPadBar.size)
numpad:SetPoint('TOPLEFT', UIParent, 'TOPLEFT', NumPadBar.left, NumPadBar.top);
numpad:SetClampedToScreen(true)
numpad:Hide()

local keyboard = NPB_KEYBOARDS[shape]
for keynum, key in ipairs(keyboard) do
	if (key.type == 'Numeric' or key.type == enter or key.type == cursor) then
		local keyname = shape .. enter .. cursor .. keynum		-- make a unique name for each button
		NPB_CreateButton(numpad, class, keyname, keynum, key.col, key.row)
		end
	end

NPB_SetState(numpad)
NPB_numpads[shape .. enter .. cursor] = numpad
end

----------------------------------------------------------------------------------------------------
function NPB_CreateButton(numpad, class, keyname, keynum, col, row)
-- creates a button based on key and attaches it to numpad
-- numpad is a frame, the parent of all the buttons to be created
-- keyname is unique, made from shape, enter pref, cursor pref, and keynum
-- keynum is the index of the key in the list of slots, 1 is first, 9 is ninth, 0 is tenth
-- key.col is the horizontal position of the key, 1 is leftmost (in drag area), 0 is one to the left, 2 is one to the right
-- key.row is the vertical position of the key, 1 is top (in drag area), 2 is one row down
----------------------------------------------------------------------------------------------------

	function NPB_DoDragStart()
	if (UnitAffectingCombat('player') == 1 or LOCK_ACTIONBAR == '1') then return end
	PickupAction(ActionButton_GetPagedID(this))
	ActionButton_UpdateHotkeys(this.buttonType)
	ActionButton_UpdateState()
	ActionButton_UpdateFlash()
	end

local button = CreateFrame('CheckButton', keyname, numpad, 'ActionBarButtonTemplate')
	-- 'frameType', 'frameName', parentFrame[, 'inheritsFrame']

local action = NPB_CLASS_SLOTS[class][keynum]
button:SetAttribute('type', 'action')
button:SetAttribute('action', action)

local stances = NPB_CLASS_STANCEBARS[class]
if (stances) then
	button:SetAttribute('*action', action) end
for sx = 0, 3 do
	if (stances and stances[sx]) then
		local stance = (keynum < 12) and stances[sx][keynum] or action
		button:SetAttribute('*action-S' .. sx, stance)
		end
	end
	
local statebutton = NPB_CLASS_STATES[class]
if (statebutton) then
	button:SetAttribute('statebutton', statebutton) end

local left = NPB_BUTTON_SIZE * (col - 1)
local top = - (NPB_BUTTON_SIZE * (row - 1) + NPB_DRAG_HANDLE)
button:SetScript('OnAttributeChanged', ActionButton_Update)		-- may not be needed in 2.1
button:SetScript('OnDragStart', NPB_DoDragStart)
button:SetPoint('TOPLEFT', left, top)
button.showgrid = 1		-- display all buttons including empty buttons (buttons with no action)
button:Show()		-- needed to show empty buttons even when showgrid=1, WoW bug?
numpad:SetAttribute('addchild', button)
end

----------------------------------------------------------------------------------------------------
function NPB_GetNumpad()
-- given shape, enter, cursor: get numpad
-- create it if it does not exist
----------------------------------------------------------------------------------------------------

local shape = NumPadBar.shape
local enter = NumPadBar.enter
local cursor = NumPadBar.cursor
if (not shape) then shape = 'Windows' end
if (not enter) then enter = '' end
if (not cursor) then cursor = '' end
local numpad = NPB_numpads[shape .. enter .. cursor]
if (numpad) then return numpad end
local localizedClass, englishClass = UnitClass('player')
local class = string.lower(englishClass)
NPB_CreateNumpad(class, shape, enter, cursor)
local numpad = NPB_numpads[shape .. enter .. cursor]
return numpad
end

----------------------------------------------------------------------------------------------------
function NPB_ShowNumpad()
----------------------------------------------------------------------------------------------------

local numpad = NPB_GetNumpad()
local keyboard = NPB_KEYBOARDS[NumPadBar.shape]
ClearOverrideBindings(numpad)
for keynum, key in ipairs(keyboard) do
	if (key.type == 'Numeric' or key.type == NumPadBar.enter or key.type == NumPadBar.cursor) then
		local name = NumPadBar.shape  .. NumPadBar.enter .. NumPadBar.cursor .. keynum
			-- this is the unique name of the button
		local label = (NumPadBar.labels) and key.label or nil
		getglobal(name .. 'HotKey'):SetText(label)
		SetOverrideBindingClick(numpad, false, key.binding, name)
		SetOverrideBindingClick(numpad, false, 'Shift-' .. key.binding, name)
			-- key.binding is the key binding such as A or Enter or NumpadDecimal
			-- also disables the default WoW key bindings for shift-numpad functions
			-- especially Previous Action Bar and Next Action Bar
			-- so that your bottom bar is not accidentally changed in combat
			-- SetOverrideBindingClick(owner, isPriority, 'KEY', 'buttonName' [,'mouseButton'])
			-- only works when the button is showing (according to blizzard)
		end
	end
numpad:SetScale(NumPadBar.size)
numpad:SetPoint('TOPLEFT', UIParent, 'TOPLEFT', NumPadBar.left, NumPadBar.top);
if (not NumPadBar.leftclicked or not NumPadBar.rightclicked) then
	SetBackdrop(numpad) end	-- if user has not moved or used the menu, display the background
numpad:Show()
NumPadBar.hidden = false
end

----------------------------------------------------------------------------------------------------
function NPB_HideNumpad()
----------------------------------------------------------------------------------------------------

local numpad = NPB_GetNumpad()
ClearOverrideBindings(numpad)
numpad:Hide()
NumPadBar.hidden = true
end

----------------------------------------------------------------------------------------------------
function NPB_ShapeNumpad(shape)
----------------------------------------------------------------------------------------------------

if (not NumPadBar.shape or not NPB_KEYBOARDS[NumPadBar.shape]) then
	NumPadBar.shape = 'Windows' end
if (not shape) then  -- if not specified, then next in list after current shape
	shape = next(NPB_KEYBOARDS, NumPadBar.shape) end
if (not shape) then  -- if not specified and nothing next in list, use first in list
	shape = next(NPB_KEYBOARDS, nil) end
NPB_HideNumpad()
NumPadBar.shape = shape
NPB_ShowNumpad()
end

----------------------------------------------------------------------------------------------------
function NPB_SizeNumpad(size)
----------------------------------------------------------------------------------------------------

if (not NumPadBar.size or not NPB_SIZES[NumPadBar.size]) then
	NumPadBar.size = 1 end
if (not size) then
	size = NPB_SIZES[NumPadBar.size] end
if (not size) then
	size = 1 end
NumPadBar.left = NumPadBar.left * NumPadBar.size / size
NumPadBar.top = NumPadBar.top * NumPadBar.size / size
NumPadBar.size = size
NPB_ShowNumpad()
end


----------------------------------------------------------------------------------------------------
function NPB_ToggleEnter()
----------------------------------------------------------------------------------------------------

NPB_HideNumpad()
NumPadBar.enter = (NumPadBar.enter == 'Enter') and '' or 'Enter'
if (NumPadBar.enter == 'Enter') then
	UIErrorsFrame:AddMessage(NPB_ENTER_ERROR, 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME) end
NPB_ShowNumpad()
end

----------------------------------------------------------------------------------------------------
function NPB_ToggleCursor()
----------------------------------------------------------------------------------------------------

NPB_HideNumpad()
NumPadBar.cursor = (NumPadBar.cursor == 'Cursor') and '' or 'Cursor'
if (NumPadBar.cursor == 'Cursor') then
	UIErrorsFrame:AddMessage(NPB_CURSOR_ERROR, 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME) end
NPB_ShowNumpad()
end

----------------------------------------------------------------------------------------------------
function NPB_ToggleLabels()
----------------------------------------------------------------------------------------------------

NPB_HideNumpad()
NumPadBar.labels = not NumPadBar.labels
NPB_ShowNumpad()
end

----------------------------------------------------------------------------------------------------
function NPB_ToggleStancebars()
----------------------------------------------------------------------------------------------------

local localizedClass, englishClass = UnitClass('player')
local class = string.lower(englishClass)
local stances = NPB_CLASS_STANCEBARS[class]
if (not stances) then NPB_Chat(NPB_STANCELESS_CLASS) return end
NumPadBar.stancebars = not NumPadBar.stancebars
local numpad = NPB_GetNumpad()
NPB_SetState(numpad)
if (NumPadBar.stancebars) then
	NPB_Chat(NPB_USING_STANCEBARS)
else
	NPB_Chat(NPB_NOT_USING_STANCEBARS) end
end

----------------------------------------------------------------------------------------------------
-- BACKDROP FUNCTIONS
----------------------------------------------------------------------------------------------------

function SetBackdrop(frame)
NPB_backdrop = true
frame:SetBackdrop(NPB_BACKDROP)
end

function ClearBackdrop(frame)
NPB_backdrop = false
frame:SetBackdrop(nil)
end

function ToggleBackdrop(frame)
if (not NPB_backdrop) then
	SetBackdrop(frame)
else
	ClearBackdrop(frame) end
end
	
----------------------------------------------------------------------------------------------------
-- DROP DOWN MENU FUNCTIONS
----------------------------------------------------------------------------------------------------

function NPB_InitializeMenu()
for ix in ipairs(NPB_MENUS) do
	UIDropDownMenu_AddButton(NPB_MENUS[ix]) end
end

function NPB_ToggleMenu()
ToggleDropDownMenu(1, nil, NPB_MenuFrame, 'cursor')
end

function NPB_DoneMenu()
local numpad = NPB_GetNumpad()
ClearBackdrop(numpad)
end

function NPBM_windows()	NPB_ShapeNumpad('Windows'); NPB_DoneMenu(); end
function NPBM_office()		NPB_ShapeNumpad('MicrosoftOffice'); NPB_DoneMenu(); end
function NPBM_multimedia()	NPB_ShapeNumpad('NaturalMultimedia'); NPB_DoneMenu(); end
function NPBM_elite()		NPB_ShapeNumpad('NaturalElite'); NPB_DoneMenu(); end
function NPBM_macintosh()	NPB_ShapeNumpad('Macintosh'); NPB_DoneMenu(); end
function NPBM_medium()		NPB_SizeNumpad(1); NPB_DoneMenu(); end
function NPBM_small()		NPB_SizeNumpad(.8); NPB_DoneMenu(); end
function NPBM_tiny()		NPB_SizeNumpad(.67); NPB_DoneMenu(); end
function NPBM_enter()		NPB_ToggleEnter(); NPB_DoneMenu(); end
function NPBM_cursor()		NPB_ToggleCursor(); NPB_DoneMenu(); end
function NPBM_labels()		NPB_ToggleLabels(); NPB_DoneMenu(); end
function NPBM_stancebars()	NPB_ToggleStancebars(); NPB_DoneMenu(); end
function NPBM_hide()		NPB_HideNumpad(); NPB_Chat(NPB_HIDDEN); NPB_DoneMenu(); end
function NPBM_help()		StaticPopup_Show('NPB_INSTRUCTIONS'); NPB_DoneMenu(); end

NPB_MENUS = {  -- must be after drop down menu functions
	{ text = NPB_MENU.windows, func = NPBM_windows },
	{ text = NPB_MENU.office, func = NPBM_office },
	{ text = NPB_MENU.multimedia, func = NPBM_multimedia },
	{ text = NPB_MENU.elite, func = NPBM_elite },
	{ text = NPB_MENU.macintosh, func = NPBM_macintosh },
	{ text = NPB_MENU.medium, func = NPBM_medium },
	{ text = NPB_MENU.small, func = NPBM_small },
	{ text = NPB_MENU.tiny, func = NPBM_tiny },
	{ text = NPB_MENU.enter, func = NPBM_enter },
	{ text = NPB_MENU.cursor, func = NPBM_cursor },
	{ text = NPB_MENU.labels, func = NPBM_labels },
	{ text = NPB_MENU.stancebars, func = NPBM_stancebars },
	{ text = NPB_MENU.hide, func = NPBM_hide },
	{ text = NPB_MENU.help, func = NPBM_help }
}

